-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local ILVL_PATTERN        = "^"..string.gsub(_G.ITEM_LEVEL, "%%d", "(%%d+)");
local LOWLEVEL_MULTIPLIER = 0.8;
local MAX_PET_LEVEL       = 25;
local MAX_ACTIVE_PETS     = 3;

-- Names of the equipment slots
local equipSlotNames =
{
  [INVSLOT_HEAD    ] = "Head",
  [INVSLOT_NECK    ] = "Neck",
  [INVSLOT_SHOULDER] = "Shoulder",
  [INVSLOT_CHEST   ] = "Chest",
  [INVSLOT_WAIST   ] = "Waist",
  [INVSLOT_LEGS    ] = "Legs",
  [INVSLOT_FEET    ] = "Feet",
  [INVSLOT_WRIST   ] = "Wrist",
  [INVSLOT_HAND    ] = "Hands",
  [INVSLOT_FINGER1 ] = "Finger0",
  [INVSLOT_FINGER2 ] = "Finger1",
  [INVSLOT_TRINKET1] = "Trinket0",
  [INVSLOT_TRINKET2] = "Trinket1",
  [INVSLOT_BACK    ] = "Back",
  [INVSLOT_MAINHAND] = "MainHand",
  [INVSLOT_OFFHAND ] = "SecondaryHand",
};

-- Auction sub-categories for 2H items that aren't actually flagged as 2H
local equippedTwoHanded =
{
  ["Bows"     ] = true,
  ["Crossbows"] = true,
  ["Guns"]      = true,
};

-- Item IDs for all know lances
local tournamentLances =
{
  [46069] = true, -- Alliance Lance
  [46070] = true, -- Horde Lance
  [46106] = true, -- Argent Lance
  [52716] = true  -- Twilight Firelance
};

-- Info for Pet Battle items that grant an XP bonus
local petBattleItems =
{
  {toyId  =  92738, buffId = 158486}, -- Safari Hat
  {itemId =  98114, buffId = 142205}, -- Pet Treat
  {itemId =  98112, buffId = 142204}, -- Lesser Pet Treat
};

-- Info for the equipment set specialization frames
local equipSetSpecInfo =
{
  dialogHeight = 0,
  textures     = {},
  fontStrings  = {},
  buttons      = {},
  setName      = nil
};

local emPopupItemInfo = {};

local validSpecs = {};

local paperDollInfo =
{
  ["Character"] =
  {
    indicators = {},
    stats =
    {
      average = 0,
      count   = 0,
      lowIlvl = 0,
      empty   = 0,
    }
  },
  ["Inspect"] =
  {
    indicators = {},
    stats =
    {
      average = 0,
      count   = 0,
      lowIlvl = 0,
      empty   = 0,
    },
    ilvlFrame = nil,
  },
};

local itemSlotRemap =
{
  -- Legion Artifacts
  [128293] = INVSLOT_MAINHAND, -- Death Knight, Frost
  [127830] = INVSLOT_MAINHAND, -- Demon Hunter, Havoc
  [128831] = INVSLOT_MAINHAND, -- Demon Hunter, Vengeance
  [128859] = INVSLOT_MAINHAND, -- Druid, Feral
  [128822] = INVSLOT_MAINHAND, -- Druid, Guardian
  [133959] = INVSLOT_MAINHAND, -- Mage, Fire
  [133948] = INVSLOT_MAINHAND, -- Monk, Windwalker
  [128866] = INVSLOT_OFFHAND,  -- Paladin, Protection
  [133958] = INVSLOT_MAINHAND, -- Priest, Shadow
  [128869] = INVSLOT_MAINHAND, -- Rogue, Assassination
  [134552] = INVSLOT_MAINHAND, -- Rogue, Outlaw
  [128479] = INVSLOT_MAINHAND, -- Rogue, Subtlety
  [128936] = INVSLOT_MAINHAND, -- Shaman, Elemental
  [128873] = INVSLOT_MAINHAND, -- Shaman, Enhancement
  [128934] = INVSLOT_MAINHAND, -- Shaman, Restoration
  [137246] = INVSLOT_OFFHAND,  -- Warlock, Demonology
  [134553] = INVSLOT_MAINHAND, -- Warrior, Fury
  [128288] = INVSLOT_OFFHAND,  -- Warrior, Protection
};

-- Pet Tamer IDs
local petTamerIds =
{
  [63194] = true,
  [64330] = true,
  [65648] = true,
  [65651] = true,
  [65655] = true,
  [65656] = true,
  [66126] = true,
  [66135] = true,
  [66136] = true,
  [66137] = true,
  [66352] = true,
  [66372] = true,
  [66412] = true,
  [66422] = true,
  [66436] = true,
  [66442] = true,
  [66452] = true,
  [66466] = true,
  [66478] = true,
  [66512] = true,
  [66515] = true,
  [66518] = true,
  [66520] = true,
  [66522] = true,
  [66550] = true,
  [66551] = true,
  [66552] = true,
  [66553] = true,
  [66557] = true,
  [66635] = true,
  [66636] = true,
  [66638] = true,
  [66639] = true,
  [66675] = true,
  [66730] = true,
  [66733] = true,
  [66734] = true,
  [66738] = true,
  [66739] = true,
  [66741] = true,
  [66815] = true,
  [66819] = true,
  [66822] = true,
  [66824] = true,
  [66918] = true,
  [67370] = true,
  [68462] = true,
  [68463] = true,
  [68464] = true,
  [68465] = true,
  [71924] = true,
  [71926] = true,
  [71927] = true,
  [71929] = true,
  [71930] = true,
  [71931] = true,
  [71932] = true,
  [71933] = true,
  [71934] = true,
  [73626] = true,
  [79179] = true,
  [83837] = true,
  [85519] = true,
  [87110] = true,
  [87122] = true,
  [87123] = true,
  [87124] = true,
  [87125] = true,
  [90675] = true,
  [91014] = true,
  [91015] = true,
  [91016] = true,
  [91017] = true,
  [97804] = true,
  [98270] = true,
  [99035] = true,
  [99077] = true,
  [99150] = true,
  [99182] = true,
  [99210] = true,
  [99878] = true,
  [104553] = true,
  [104970] = true,
  [105093] = true,
  [105250] = true,
  [105323] = true,
  [105386] = true,
  [105387] = true,
  [105455] = true,
  [105512] = true,
  [105674] = true,
  [105779] = true,
  [106552] = true,
  [107489] = true,
  [115286] = true,
  [115307] = true,
  [117934] = true,
  [117950] = true,
  [117951] = true,
  [124617] = true,
};

local missingPetXPBuffs = nil;

local lockedItemInfo = {};

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

local emSetPopupFilterSets;
local emSetPopupFilterSpecs;
local getItemLinkLevel;
local petBattleEquipmentGetMissing;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to update the EM Set popup
--
local function emSetPopupListUpdate()
  -- Local Variables
  local _;
  local scrollFrame = NigglesEMSetPopupScrollFrame;
  local parent = NigglesEMSetPopup;
  local numItems = #emPopupItemInfo;
  local scrollOffset = HybridScrollFrame_GetOffset(scrollFrame);
  local buttons = scrollFrame.buttons;
  local selectedInfo = parent.selectedInfo;
  local setIcon;
  local button;

  -- Update the scroll frame's buttons
  for buttonIdx = 1, #buttons do
    -- Assign the button to a more convenient variable
    button = buttons[buttonIdx];

    -- Check if there is an equipment set for the button to display
    if (buttonIdx+scrollOffset <= numItems) then
      -- Save the info for the button
      button.info = emPopupItemInfo[buttonIdx+scrollOffset];

      -- Set the name of the equipment set the button is for
      button.text:SetText(button.info.name);

      -- Set the colour of the button's text
      button.text:SetTextColor(button.info.textColor.r,
        button.info.textColor.g, button.info.textColor.b);

      -- Set the button's textures
      button.icon:SetTexture(button.info.texture or
        "Interface\\Icons\\INV_Misc_QuestionMark");
      button.selected:SetShown(
        (selectedInfo ~= nil) and (button.info == selectedInfo));
      button.equipped:SetShown(button.info.isActive or false);
      if ((buttonIdx+scrollOffset) == 1) then
        button.bgTop:Show();
        button.bgMiddle:SetPoint("TOP", button.bgTop, "BOTTOM");
      else
        button.bgTop:Hide();
        button.bgMiddle:SetPoint("TOP", button, "TOP", 0, 0);
      end
      if ((buttonIdx+scrollOffset) == numItems) then
        button.bgBottom:Show();
        button.bgMiddle:SetPoint("BOTTOM", button.bgBottom, "TOP");
      else
        button.bgBottom:Hide();
        button.bgMiddle:SetPoint("BOTTOM");
      end
      button.stripe:SetShown((buttonIdx+scrollOffset)%2 == 0);

      -- Show the button
      button:Show();
    else
      button:Hide();
    end
  end

  -- Update the scroll frame
  HybridScrollFrame_Update(scrollFrame, numItems*buttons[1]:GetHeight()+6,
    scrollFrame:GetHeight());

  return;
end

--
-- Function called when the active spec changes. It works out which equipment
-- sets are for the new active spec. If there is one, that set is automatically
-- equipped. If there's more than one then a popup is displayed for the user
-- to select which set to equip.
--
local function emSetSpecChanged()
  -- Build a list of EM sets for the active spec
  emSetPopupFilterSets();

  -- Check how many sets were found for the active spec
  if (#emPopupItemInfo == 1) then
    NigglesEMSetPopup:Hide();
    NigglesEMSetPopup.processEvent = false;
    UseEquipmentSet(emPopupItemInfo[1].name);
  elseif (#emPopupItemInfo > 1) then
    -- Show the EM Set popup
    NigglesEMSetPopup.mode = "set";
    NigglesEMSetPopup:Show();
  end

  return;
end

--
-- Function to display the tooltip for an EM set spec button
--
local function emSetSpecBtnOnEnter(self)
  -- Local Variables
  local _;
  local specName = nil;
  local specDesc;

  -- Get the name and description for the button's specialization
  _, specName, specDesc = GetSpecializationInfoByID(self.specId);

  -- Display a tooltip for the button
  GameTooltip:SetOwner(self, "ANCHOR_BOTTOMRIGHT", 0);
  GameTooltip:AddLine(specName,
    HIGHLIGHT_FONT_COLOR.r,
    HIGHLIGHT_FONT_COLOR.g,
    HIGHLIGHT_FONT_COLOR.b);
  GameTooltip:AddLine(specDesc,
    NORMAL_FONT_COLOR.r,
    NORMAL_FONT_COLOR.g,
    NORMAL_FONT_COLOR.b);
  GameTooltip:SetMinimumWidth(150, true);
  GameTooltip:Show();

  return;
end

--
-- Function called when an EM set is equipped successfully. It works out which
-- specs are for the new EM set. If there is one and it isn't the active spec,
-- that spec is automatically equipped. If there is more than one then a popup
-- is displayed for the user to select which spec to activate.
--
--
local function equipSetEquipped(setName)
  -- Local Variables
  local activeSpecId = validSpecs[GetSpecialization(false, false)];

  -- Build a list of specs for the active EM set
  emSetPopupFilterSpecs(setName);

  -- Check how many specs were found for the EM set
  if ((#emPopupItemInfo == 1) and
      (emPopupItemInfo[1].id ~= activeSpecId)) then
    NigglesEMSetPopup:Hide();
    NigglesEMSetPopup.processEvent = false;
    SetSpecialization(validSpecs[emPopupItemInfo[1].id], false);
  elseif (#emPopupItemInfo > 1) then
    -- Show the EM Set popup
    NigglesEMSetPopup.mode = "spec";
    NigglesEMSetPopup:Show();
  end

  return;
end

--
-- Function to build a list of equipment sets marked as being for the active
-- specialization.
--
emSetPopupFilterSets = function()
  -- Local Variables
  local _;
  local activeSpecId = validSpecs[GetSpecialization(false, false)];
  local equipSetSpecs = NigglesDBC.equipSetSpec;
  local numFiltered = 0;
  local selectedValid = false;
  local selectedInfo = NigglesEMSetPopup.selectedInfo;
  local numSets = GetNumEquipmentSets();
  local setName;
  local setTexture;
  local isEquipped;
  local missingItems;

  -- Clear any existing item info
  wipe(emPopupItemInfo);

  -- Check there is an active spec
  if (activeSpecId ~= nil) then
    -- Filter the EM sets
    -- NOTE: 'GetEquipmentSetInfo()' must be used to show in Blizzard UI order
    for idx = 1, numSets do
      -- Get the name of the EM set
      setName, setTexture, _, isEquipped, _, _, _, missingItems =
        GetEquipmentSetInfo(idx);

      -- Check EM set is associated with the active spec
      if ((equipSetSpecs[setName] ~= nil) and
          (equipSetSpecs[setName][activeSpecId] ~= nil)) then

        -- Add the EM set to the filtered EM sets
        emPopupItemInfo[#emPopupItemInfo+1] =
        {
          name      = setName,
          texture   = setTexture,
          textColor = ((missingItems > 0) and
            RED_FONT_COLOR or NORMAL_FONT_COLOR),
          isActive  = isEquipped,
        };

        -- Check if the current set is the selected set
        if ((selectedInfo ~= nil) and (selectedInfo.name == setName)) then
          selectedValid = true;
        end
      end
    end
  end

  -- Check if the selection is no longer valid
  if (not selectedValid) then
    NigglesEMSetPopup.selectedInfo = nil;
    NigglesEMSetPopup.okay:Disable();
  end

  -- Update the Equip Set popup
  if (NigglesEMSetPopup:IsShown()) then
    emSetPopupListUpdate();
  end

  return;
end

--
-- Function to build a list of specs marked as being for the equipment set
-- just equipped
--
emSetPopupFilterSpecs = function(setName)
  -- Local Variables
  local _;
  local activeSpecId = validSpecs[GetSpecialization(false, false)];
  local emSetInfo = NigglesDBC.equipSetSpec[setName];
  local numSpecs = GetNumSpecializations(false, false);
  local selectedValid = false;
  local specId;
  local specName;
  local specTexture;

  -- Clear any existing item info
  wipe(emPopupItemInfo);

  -- Check a valid equipment set was loaded
  if (type(emSetInfo) == "table") then
    -- Filter the specs
    for specIdx = 1, numSpecs do
      -- Check if the spec is assigned to the EM set
      if ((validSpecs[specIdx] ~= nil) and
          (emSetInfo[validSpecs[specIdx]] ~= nil)) then
        -- Get the info for the spec
        specId, specName, _, specTexture = GetSpecializationInfo(specIdx);

        -- Add the spec to the filtered EM sets
        emPopupItemInfo[#emPopupItemInfo+1] =
        {
          name      = specName,
          texture   = specTexture,
          textColor = NORMAL_FONT_COLOR,
          isActive  = (specId == activeSpecId),
          id        = specId,
        };

        -- Check if the current set is the selected set
        if ((selectedInfo ~= nil) and (selectedInfo.id == specId)) then
          selectedValid = true;
        end
      end
    end
  end

  -- Check if the selection is no longer valid
  if (not selectedValid) then
    NigglesEMSetPopup.selectedInfo = nil;
    NigglesEMSetPopup.okay:Disable();
  end

  -- Update the Equip Set popup
  if (NigglesEMSetPopup:IsShown()) then
    emSetPopupListUpdate();
  end

  return;
end

--
-- Function to scan all equipped items. It works out an average ilvl, as well
-- as counting the number of empty slots.
--
local function equipmentScan(target)
  -- Local Variables
  local _;
  local equippedIlvl = {};
  local indicators = paperDollInfo[target].indicators;
  local infoAvailable = true;
  local itemLink;
  local itemQuality;
  local itemSubCat;
  local itemType;
  local itemLvl;
  local lowIlvl;
  local stats = paperDollInfo[target].stats;
  local totalIlvl = 0;
  local totalIlvl = 0;
  local unit;

  -- Initialise some variables
  unit = ((target == "Inspect") and InspectFrame.unit or "player");

  -- Reset the stats for the equipped items
  stats.average = 0;
  stats.count   = 0;
  stats.lowIlvl = 0;
  stats.empty   = 0;

  -- Scan all equipment slots
  for slotId, slotName in pairs(equipSlotNames) do
    -- Initialise some variables
    equippedIlvl[slotId] = 0;

    -- Check if the slot is contains an item
    itemLink = GetInventoryItemLink(unit, slotId);
    if (itemLink ~= nil) then
      -- Check if the item's level should be taken from another slot
      -- NOTE: This is to deal with a bug in the Blizzard API that means
      --       Legion artifact weapons that use both the main and off-hand
      --       slots will always give an item level of 750 for one of them.
      _, _, itemId = string.find(itemLink, "|Hitem:(%d+):");
      if (itemId ~= nil) then
        itemId = tonumber(itemId);
        if (itemSlotRemap[itemId] ~= nil) then
          itemLink = GetInventoryItemLink(unit, itemSlotRemap[itemId]);
        end
      end

      -- Get the item's info
      _, _, itemQuality = GetItemInfo(itemLink);
      itemLvl           = getItemLinkLevel(itemLink);
      if ((itemQuality ~= nil) and (itemLvl ~= nil)) then
        -- Update the info for calculating the average item level
        equippedIlvl[slotId] = itemLvl;
        totalIlvl            = totalIlvl+itemLvl;
        stats.count          = stats.count+1;
      else
        infoAvailable = false;
        break;
      end
    end
  end

  -- Check item info could be obtained for all items
  if (infoAvailable) then
    -- Check if the info needs adjusting to account for 2H weapons
    -- NOTE: The Blizzard API average item level calculation treats 2H weapons
    --       as if there was a 2nd weapon in the off-hand of same item level.
    if ((equippedIlvl[INVSLOT_OFFHAND] == 0) and
        (equippedIlvl[INVSLOT_MAINHAND] > 0)) then
      -- Get the main hand's type and sub category
      _, _, _, _, _, _, itemSubCat, _, itemType =
        GetItemInfo(GetInventoryItemLink(unit, INVSLOT_MAINHAND));
      if ((itemType == "INVTYPE_2HWEAPON") or
          (equippedTwoHanded[itemSubCat] ~= nil)) then
        -- Set the off-hand to the same item level as the main hand
        equippedIlvl[INVSLOT_OFFHAND] = equippedIlvl[INVSLOT_MAINHAND];

        -- Update the info for calculating the average item level
        totalIlvl   = totalIlvl+equippedIlvl[INVSLOT_MAINHAND];
        stats.count = stats.count+1;
      end
    end

    -- Work out the average item level
    if (stats.count > 0) then
      stats.average = totalIlvl/stats.count;
    else
      stats.average = 0;
    end

    -- Work out how many low level items are equipped
    lowIlvl = stats.average*LOWLEVEL_MULTIPLIER;
    for slotId, slotIlvl in pairs(equippedIlvl) do
      if (slotIlvl == 0) then
        stats.empty = stats.empty+bit.lshift(1, slotId);
      elseif (slotIlvl < lowIlvl) then
        stats.lowIlvl = stats.lowIlvl+bit.lshift(1, slotId);
      end
    end
  else
    -- Reset the stats for the equipped items
    stats.count   = 0;
    stats.lowIlvl = 0;
    stats.empty   = 0;

    -- Scan the equipment again in a short while
    C_Timer.After(0.5,
      function()
        equipmentScan(target);
      end);
  end

  -- Update the low level item indicators
  if ((NigglesDBC.settings.UiLowLevelItems) and
      (stats.lowIlvl > 0)) then
    lowIlvl = stats.average*LOWLEVEL_MULTIPLIER;
    for slotId, slotIlvl in pairs(equippedIlvl) do
      if ((slotIlvl ~= 0) and
          (slotIlvl < lowIlvl)) then
        -- Create the equipment slot's low Ilvl indicator, if required
        -- NOTE: A font string can't be created as a child of the equipment
        --       slot because it can't be placed above the cooldown.
        if (indicators[slotId] == nil) then
          indicators[slotId] = CreateFrame("Frame", nil,
            _G[target..equipSlotNames[slotId].."Slot"],
            "NigglesLowIlvlTemplate");
        end
        indicators[slotId]:Show();
      elseif (indicators[slotId] ~= nil) then
        indicators[slotId]:Hide();
      end
    end
  else
    -- Hide all the low ilvl indicators
    for _, indicator in pairs(indicators) do
      indicator:Hide();
    end
  end

  return;
end

--
-- Function called to process events for the EM Set popup
--
local function equipSetPopupOnEvent(self, event, ...)
  -- Local Variables
  local _;

  -- Process the event
  if (event == "UNIT_SPELLCAST_SUCCEEDED") then
    local _, _, _, _, spellId = ...;
    if ((spellId == 200749) and (self.processEvent)) then
      emSetSpecChanged();
    end
  elseif (event == "UNIT_SPELLCAST_STOP") then
    self.processEvent = true;
  elseif (event == "EQUIPMENT_SWAP_FINISHED") then
    local wasSuccessful, setName = ...;
    if ((wasSuccessful) and (setName ~= nil) and (self.processEvent)) then
      equipSetEquipped(setName);
    else
      self.processEvent = true;
    end
  elseif (event == "EQUIPMENT_SETS_CHANGED") then
    if (NigglesEMSetPopup:IsShown()) then
      emSetPopupFilterSets();
    end
  elseif (event == "PLAYER_EQUIPMENT_CHANGED") then
    if ((NigglesDBC.settings.UiLowLevelItems) or
        (NigglesDBC.settings.WarningLowLevelItems)) then
      equipmentScan("Character");
    end
    if ((NigglesEMSetPopup:IsShown()) and
        (NigglesEMSetPopup.mode == "set")) then
      emSetPopupFilterSets();
    end
  end

  return;
end

--
-- Function called when the EM Set popup is shown
--
local function equipSetPopupOnShow(self)
  -- Local Variables
  local okayLabel = "";
  local title = "";

  -- Clear any old selection
  self.okay:Disable();
  self.selectedInfo = nil;

  -- Set the title and label for the OK button
  if (self.mode == "spec") then
    title     = L["ActivateSpec"];
    okayLabel = TALENT_SPEC_ACTIVATE;
  elseif (self.mode == "set") then
    title     = L["EquipSet"];
    okayLabel = EQUIPSET_EQUIP;
  end
  NigglesEMSetPopup.title:SetText(title);
  NigglesEMSetPopup.okay:SetText(okayLabel);

  -- Update the scroll frame
  emSetPopupListUpdate();

  return;
end

--
-- Function called when the 'OK' button is clicked in the EM Set dialog
--
local function equipSetSpecOkayOnClick(self, mouseBtn)
  -- Check if the EM Set spec feature is enabled
  if (NigglesDBC.settings.UiEMSetSpec) then
    -- Local Variables
    local _;
    local equipSetSpecs = NigglesDBC.equipSetSpec;
    local origName = GearManagerDialogPopup.origName;
    local setName = equipSetSpecInfo.setName;
    local buttonSpecs = {};

    -- Check a name has been specified for the EM set
    if (setName ~= nil) then
      -- Check if the EM set has been renamed
      if ((origName ~= nil) and
          (origName ~= equipSetSpecInfo.name) and
          (equipSetSpecs[origName] ~= nil)) then
        -- Delete the EM spec's old info
        equipSetSpecs[origName] = nil;
      end

      -- Work out which specs are selected
      for _, button in ipairs(equipSetSpecInfo.buttons) do
        if (button:GetChecked()) then
          buttonSpecs[button.specId] = 1;
        end
      end

      -- If there are some specs are selected...
      if (next(buttonSpecs) ~= nil) then
        -- ...save them for the EM set
        equipSetSpecs[setName] = buttonSpecs;
      end
    end

    -- Refresh the EM Set popup, if required
    if (NigglesEMSetPopup:IsShown()) then
      emSetPopupFilterSets();
    end
  end

  return;
end

--
-- Function to pre-process a click on the 'OK' button in the EM Set dialog.
-- It saves the name of the EM set. This function is required because the
-- 'OnClick' function for the button clears the name from both it's internal
-- variable and the edit box in the dialog.
--
local function equipSetSpecOkayPreClick(self, mouseBtn)
  -- Save the name of the EM set
  equipSetSpecInfo.setName = GearManagerDialogPopup.name;

  return;
end

--
-- Function called when 'RecalculateGearManagerDialogPopup' is called by the
-- Blizzard code to set the EM set displayed in the EM Set dialog.
--
local function equipSetSpecRecalculate(setName, setIcon)
  -- Local Variables
  local _;
  local buttons  = equipSetSpecInfo.buttons;
  local setSpecs = NigglesDBC.equipSetSpec[setName];

  -- Check if the EM Set spec feature is enabled
  if (NigglesDBC.settings.UiEMSetSpec) then
    -- Initialise the buttons for each spec
    for _, button in ipairs(equipSetSpecInfo.buttons) do
      -- Assign the button to a more convenient variable
      button:SetChecked(((setSpecs ~= nil) and
        (setSpecs[button.specId] ~= nil)));
    end
  end

  return;
end

--
-- Function to get the item level for an item from a tooltip.
-- NOTE: This function is required because 'GetItemInfo()' and even
--       'GetDetailedItemLevelInfo()' do not return reliable item levels,
--        especially for items with modifiers such as 'Mythic', 'Titanforged'
--        or Timewalking.
--
getItemLinkLevel = function(itemLink)
  -- Local Variables
  local _;
  local itemIlvl = nil;
  local itemName
  local itemQuality;
  local textLeft = NigglesTooltip.textLeft;
  local numLines

  -- Set the tooltip and scan its lines for the item level
  numLines = NigglesTooltip:SetContents("SetHyperlink", itemLink);
  for lineIdx = 2, math.min(numLines, 5) do
    _, _, itemIlvl = string.find(textLeft[lineIdx]:GetText(), ILVL_PATTERN);
    if (itemIlvl ~= nil) then
      itemIlvl = tonumber(itemIlvl);
      break;
    end
  end

  return itemIlvl;
end

--
-- Function called when the Gossip panel is hidden. If a warning about missing
-- pet XP buffs is displayed, the current time is saved. The time is used to
-- determine if a warning should be displayed when a pet battle starts.
--
local function gossipOnHide()
  -- Check if the warning about missing pet XP buffs exists
  if (missingPetXPBuffs ~= nil) then
    -- Check if the warning about missing pet XP buffs is shown
    if (missingPetXPBuffs:IsShown()) then
      missingPetXPBuffs.time = GetTime();
    end

    -- Hide the warning about missing pet XP buffs
    missingPetXPBuffs:Hide();
  end

  return;
end

--
-- Function called when the options are changed in the Gossip frame
--
local function gossipOptionsOnUpdate(...)
  -- Local Variables
  local buttonIdx = NUMGOSSIPBUTTONS;
  local child;
  local frame = missingPetXPBuffs;
  local missingEquipment;

  -- Check if the current NPC is a pet tamer
  if ((NigglesDBC.settings.WarningPetBattle) and
      (L.npcIsPetTamer())) then
    missingEquipment = petBattleEquipmentGetMissing(true);
    if (missingEquipment ~= "") then
      -- Create the missing equipment frame, if required
      if (frame == nil) then
        frame = CreateFrame("FRAME", nil, GossipGreetingScrollChildFrame);
        frame:SetSize(1,1);
        frame:SetPoint("LEFT", 8, 0);
        frame:SetPoint("RIGHT", GossipGreetingText, "RIGHT");
        missingPetXPBuffs = frame;
        child = frame:CreateTexture(nil, "ARTWORK", nil);
        child:SetTexture("Interface\\DialogFrame\\UI-Dialog-Icon-AlertNew");
        child:SetPoint("TOPLEFT");
        child:SetSize(40, 40);
        frame.icon = child;
        child = frame:CreateFontString(nil, "ARTWORK");
        child:SetFontObject(GossipTitleButton1:GetNormalFontObject());
        child:SetJustifyH("LEFT");
        child:SetPoint("TOPRIGHT", 0, -8);
        child:SetPoint("LEFT", frame.icon, "RIGHT", 8, 0);
        child:SetText(L["PetBattleMissingBuffs"]);
        frame.label = child;
        child = frame:CreateFontString(nil, "ARTWORK");
        child:SetFontObject(GossipTitleButton1:GetNormalFontObject());
        child:SetJustifyH("LEFT");
        child:SetPoint("TOPLEFT", frame.label, "BOTTOMLEFT", 12, -8);
        child:SetPoint("RIGHT");
        frame.list = child;
      end

      -- Find the last visible gossip title
      while ((buttonIdx > 0) and
             (not _G["GossipTitleButton"..buttonIdx]:IsShown())) do
        buttonIdx = buttonIdx-1;
      end

      -- Show the missing equipment frame
      frame.list:SetText(missingEquipment);
      frame:SetPoint("TOP", _G["GossipTitleButton"..buttonIdx], "BOTTOM",
        0, -_G["GossipTitleButton"..buttonIdx]:GetHeight());
      frame:Show();

      -- Save the GUID of the NPC
      missingPetXPBuffs.npcGuid = UnitGUID("npc");
    end
  elseif (frame ~= nil) then
    frame:Hide();
  end

  return;
end

--
-- Function to update the item level of the inspected unit
--
local function inspectedUnitUpdateIlvl()
  -- Local Variables
  local ilvlFrame = paperDollInfo["Inspect"].ilvlFrame;
  
  -- Check if the inspected character's item level should be displayed
  if ((NigglesDBC.settings.UiInspectedItemLevel) and
      (InspectFrame.unit ~= nil) and
      (UnitLevel(InspectFrame.unit) >=
        MIN_PLAYER_LEVEL_FOR_ITEM_LEVEL_DISPLAY)) then
    -- Scan the inspected character's equipment and calculate the item level
    equipmentScan("Inspect");

    -- Update the item level displayed
    ilvlFrame.text:SetText(
      STAT_AVERAGE_ITEM_LEVEL..": "..HIGHLIGHT_FONT_COLOR_CODE..
      math.ceil(paperDollInfo["Inspect"].stats.average));
    ilvlFrame:Show();
  elseif (ilvlFrame ~= nil) then
    ilvlFrame:Hide();
  end

  return;
end

--
-- Function called when the inspected unit's inventory changes
--
local function inspectedUnitOnEvent(self, event, unit)
  -- Check if the event should be processed
  if ((NigglesDBC.settings.UiInspectedItemLevel) and
      (InspectFrame:IsVisible())) then
    -- Process the event
    if (((event == "UNIT_INVENTORY_CHANGED") and
         (unit == InspectFrame.unit)) or
        (event == "GET_ITEM_INFO_RECEIVED")) then
      inspectedUnitUpdateIlvl();
    end
  end

  return;
end

--
-- Function to check if any appearance an item has is known. If the item
-- doesn't have an appearance then it is considered known.
--
local function isAppearanceKnown(itemId)
  -- Local Variables
  local _;
  local appearanceId = C_TransmogCollection.GetItemInfo(itemId);
  local isKnown = false;
  local sourceIds;
  local sourceId;
  local sourceInfo;

  -- Check if the item has an appearance
  if (appearanceId ~= null) then
    -- Get the source IDs for the item's appearance
    sourceIds = C_TransmogCollection.GetAllAppearanceSources(appearanceId);
    if (sourceIds ~= nil) then
      for _, sourceId in ipairs(sourceIds) do
        sourceInfo = C_TransmogCollection.GetSourceInfo(sourceId);
        if ((sourceInfo ~= nil) and (sourceInfo.isCollected)) then
          isKnown = true;
          break;
        end
      end
    end
  else
    isKnown = true;
  end

  return isKnown;
end

--
-- Function to process a button click in a warning popup for missing pet battle
-- XP buffs
--
local function petBattleBuffsOnClick(popup, id)
  -- Check if the 'Forfeit' button was clicked
  if (id == 2) then
    C_PetBattles.ForfeitGame();
  end

  -- Hide the popup
  StaticPopupSpecial_Hide(popup);

  return;
end

--
-- Function to check for Pet Battle equipment that isn't equipped, if there
-- are any pets that aren't maximum level.
--
-- NOTE: This function will need enhancement if Blizzard ever add items for
--       ring, trinket, main or off-hand slots.
--
petBattleEquipmentGetMissing = function(simple)
  -- Local Variables
  local _;
  local forfeitPenalty;
  local itemLink;
  local itemName;
  local itemTexture;
  local nonMaxLevel = false;
  local petGUID;
  local petLevel;
  local string = "";
  local idx;

  -- Work out if there are any pets that aren't max level
  for idx = 1, MAX_ACTIVE_PETS do
    petGUID = C_PetJournal.GetPetLoadOutInfo(idx);
    if (petGUID ~= nil) then
      _, _, petLevel = C_PetJournal.GetPetInfoByPetID(petGUID);
      if (petLevel < MAX_PET_LEVEL) then
        nonMaxLevel = true;
        break;
      end
    end
  end

  -- Check if there are any pets that aren't max level
  if (nonMaxLevel) then
    -- Get the localized names for the pet battle spells, if required
    if (petBattleItems[1].buffName == nil) then
      for _, itemInfo in ipairs(petBattleItems) do
        itemInfo.buffName = GetSpellInfo(itemInfo.buffId);
      end
    end

    -- Check if there are any XP buffs missing that are available
    for itemIdx, itemInfo in ipairs(petBattleItems) do
      if (UnitBuff("player", itemInfo.buffName) == nil) then
        -- Get info for the missing XP buff
        itemLink    = nil;
        itemTexture = nil;
        if ((itemInfo.itemId ~= nil) and
            (GetItemCount(itemInfo.itemId, false) > 0)) then
          itemName, itemLink, _, _, _, _, _, _, _, itemTexture =
            GetItemInfo(itemInfo.itemId);
        elseif ((itemInfo.toyId ~= nil) and
                (PlayerHasToy(itemInfo.toyId))) then
          _, itemName, itemTexture = C_ToyBox.GetToyInfo(itemInfo.toyId);
          itemLink = C_ToyBox.GetToyLink(itemInfo.toyId);
        end

        -- Check the info for the missing XP buff could be obtained
        if ((itemTexture ~= nil) and (itemLink ~= nil)) then
          if (simple) then
            string = string..itemName.."\n";
          else
            -- NOTE: The '|r' is done after the '\n' to overcome a Blizzard bug
            string = string.."\124T"..itemTexture..":0:0:0:-8\124t "..
              itemLink.."\n|r";
          end
        end
      end
    end
  end

  return string;
end

-------------------------------------------------------------------------------
--                 A  D  D  O  N     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to initialise the EM set spec. It adjusts the points on various
-- frames in the EM Set panel to better accommodate additional frames.
--
function L.emSetSpecInit()
  -- Local Variables
  local _;
  local child;
  local childText;
  local childType;
  local dialog = GearManagerDialogPopup;
  local equipSetSpecs = NigglesDBC.equipSetSpec;
  local iconButton = GearManagerDialogPopupButton1;
  local iconLabel = GearManagerDialogPopup.BorderBox.ChooseIconText;
  local numValidSpecs = GetNumSpecializations(false, false);
  local scrollFrame = GearManagerDialogPopupScrollFrame;
  local specId;
  local xOffset;
  local yOffset;

  -- Save the original height of the dialog
  equipSetSpecInfo.dialogHeight = dialog:GetHeight();

  -- Reposition the scroll frame
  scrollFrame:ClearAllPoints();
  scrollFrame:SetPoint("BOTTOMRIGHT", GearManagerDialogPopupCancel,
    "TOPRIGHT", -28, 4);

  -- Reposition the icon label
  iconLabel:ClearAllPoints();
  iconLabel:SetPoint("TOPLEFT", scrollFrame, "TOPLEFT", 26, 0);

  -- Reposition all the icon buttons
  iconButton:ClearAllPoints();
  iconButton:SetPoint("TOPLEFT", scrollFrame, "TOPLEFT",
    26, -(iconLabel:GetHeight()+5));

  -- Initialise the list of valid specs
  for idx = 1, numValidSpecs do
    specId = GetSpecializationInfo(idx, false, false);
    validSpecs[idx] = specId;
    validSpecs[specId] = idx;
  end

  -- Validate the EM Set specs
  for setName, setSpecs in pairs(equipSetSpecs) do
    if (type(setSpecs) == "table") then
      -- Check all the specs set for the EM Set are valid
      for specId, _ in pairs (setSpecs) do
        if (validSpecs[specId] == nil) then
          setSpecs[specId] = nil;
        end
      end

      -- If there are no specs left for the EM Set...
      if (next(setSpecs) == nil) then
        -- ...clear the EM set
        equipSetSpecs[setName] = nil;
      end
    else
      equipSetSpecs[setName] = nil;
    end
  end

  return;
end

--
-- Function to check for Pet Battle items aren't equipped at start of a battle.
--
function L.equipmentCheckPetBattleXPBuffs()
  -- Local Variables
  local _;
  local forfeitPenalty;
  local missingEquipment;

  -- Check if there are any missing buffs
  if ((missingPetXPBuffs == nil) or
      (missingPetXPBuffs.time == nil) or
      (GetTime()-missingPetXPBuffs.time > 2)) then
    missingEquipment = petBattleEquipmentGetMissing();
    if (missingEquipment ~= "") then
      -- Display a warning about the missing buffs
      forfeitPenalty = C_PetBattles.GetForfeitPenalty();
      NigglesWarningPopupInit(L["PetBattleMissingBuffs"],
        (forfeitPenalty > 0 and
          format(L["PetBattleHealthLoss"], forfeitPenalty) or nil),
        CONTINUE, PET_BATTLE_FORFEIT, petBattleBuffsOnClick);
      NigglesWarningPopup.list:SetText(missingEquipment);
      NigglesWarningPopup.list:Show();
      StaticPopupSpecial_Show(NigglesWarningPopup);
    end
  end

  return;
end

--
-- Function to dump debug info for the equipment features
--
function L.equipmentDumpDebug()
  -- Local Variables
  local info = ChatTypeInfo["SYSTEM"];
  local key;
  local value;

  -- Dump the equipment stats for the player
  DEFAULT_CHAT_FRAME:AddMessage("    Ngl-equipment:",
    info.r, info.g, info.b);
  DEFAULT_CHAT_FRAME:AddMessage("        Player Equipment Stats:",
    info.r, info.g, info.b);
  for key, value in pairs(paperDollInfo["Character"].stats) do
    DEFAULT_CHAT_FRAME:AddMessage("            "..key..": "..(value or ""),
    info.r, info.g, info.b);
  end

  return;
end

--
-- Function to get the number of empty slots
--
function L.equipmentGetEmptySlots()
  return paperDollInfo["Character"].stats.empty;
end

--
-- Function to get the number of low level items equipped
--
function L.equipmentGetLowIlvl()
  return paperDollInfo["Character"].stats.lowIlvl;
end

--
-- Function called when the Blizzard Inspect UI addon.
--
function L.equipmentInspectUIOnLoad()
  -- Local Variables
  local frame;
  local child;

  -- Hook the function that updates the buttons in the Inspect panel
  hooksecurefunc("InspectPaperDollFrame_UpdateButtons",
    inspectedUnitUpdateIlvl);

  -- Create a frame to display the item level of the inspected unit
  frame = CreateFrame("FRAME", nil, InspectPaperDollItemsFrame);
  frame:SetSize(100, 10);
  frame:SetPoint("BOTTOMRIGHT", -14, 14);
  paperDollInfo["Inspect"].ilvlFrame = frame;
  child = frame:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
  child:SetPoint("TOPRIGHT");
  child:SetJustifyH("RIGHT");
  frame.text = child;
  frame:SetScript("OnEvent", inspectedUnitOnEvent);
  frame:RegisterEvent("GET_ITEM_INFO_RECEIVED");
  frame:RegisterEvent("UNIT_INVENTORY_CHANGED");

  return;
end

--
-- Function to check if a lance is equipped
--
function L.equipmentIsLance(itemId)
  return (tournamentLances[itemId] ~= nil);
end

--
-- Function to save info for an item in a container when it is locked. The
-- info is used when items are sold.
--
function L.equipmentSaveItemInfo(containerId, slotId)
  -- Check the locked item is in a container
  if ((containerId ~= nil) and (slotId ~= nil)) then
    lockedItemInfo.texture, _, _, lockedItemInfo.quality, _, _,
    lockedItemInfo.link, _, _, lockedItemInfo.id =
      GetContainerItemInfo(containerId, slotId);
    if (lockedItemInfo.id ~= nil) then
      _, _, lockedItemInfo.name = string.find((lockedItemInfo.link or ""),
        "|h%[(.*)%]|h|r$");
      lockedItemInfo.inEquipSet =
        GetContainerItemEquipmentSetInfo(containerId, slotId);
      lockedItemInfo.appearanceKnown =
        isAppearanceKnown(lockedItemInfo.id);
    else
      wipe(lockedItemInfo);
    end
  else
    wipe(lockedItemInfo);
  end

  return;
end

--
-- Function to check if the current NPC is a pet tamer
--
function L.npcIsPetTamer()
  -- Local Variables
  local npcId = tonumber(string.match((UnitGUID("npc") or ""),
                  "[^-]+-[^-]+-[^-]+-[^-]+-[^-]+-([^-]+)-"));

  return ((npcId ~= nil) and (petTamerIds[npcId] == true));
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to initialise the Niggles Equipment functions
--
function NigglesEquipmentInit()
  -- Local Variables
  local idx;

  -- Hook the function used to sell items
  hooksecurefunc("UseContainerItem", NigglesEquipmentOnItemSale);

  -- Add the EM Set popup to the list of frames closed by ESC
  tinsert(UISpecialFrames, "NigglesEMSetPopup");

  -- Hook the function used to update the gossip frame
  hooksecurefunc("GossipFrameOptionsUpdate", gossipOptionsOnUpdate);
  for idx = 1, NUMGOSSIPBUTTONS do
    _G["GossipTitleButton"..idx]:Hide();
  end
  GossipFrame:HookScript("OnHide", gossipOnHide);

  return;
end

--
-- Function called when an item is being sold
--
function NigglesEquipmentOnItemSale(container, slot, target)
  -- Local Variables
  local _;
  local itemId;
  local itemLink;
  local itemName;
  local itemQuality;
  local itemTexture;
  local warning = nil;

  -- Check if the item is being sold
  if ((MerchantFrame:IsShown()) and
      (MerchantFrame.selectedTab == 1) and
      (next(lockedItemInfo) ~= nil)) then
    -- Check if the locked item is part of an equipment set
    if ((NigglesDBC.settings.WarningEquipSetItem) and
        (lockedItemInfo.inEquipSet)) then
      warning = L["EquipSetItemSold"];
    end

    -- Check if the locked item's appearance is uncollected
    if ((NigglesDBC.settings.WarningUncollectedSold) and
        (lockedItemInfo.id ~= nil) and
        (not isAppearanceKnown(lockedItemInfo.id)) and
        (lockedItemInfo.quality == 2)) then
      warning = L["UncollectedAppearanceSold"];
    end
   
    -- Display a warning, if required
    if (warning ~= nil) then
      NigglesWarningPopupInit(warning, nil, OKAY, nil, nil)
      NigglesWarningPopup.item.icon:SetTexture(lockedItemInfo.texture);
      NigglesWarningPopup.item.text:SetTextColor(
        ITEM_QUALITY_COLORS[lockedItemInfo.quality].r,
        ITEM_QUALITY_COLORS[lockedItemInfo.quality].g,
        ITEM_QUALITY_COLORS[lockedItemInfo.quality].b,
        1);
      NigglesWarningPopup.item.text:SetText(lockedItemInfo.name);
      NigglesWarningPopup.item:Show();
      StaticPopupSpecial_Show(NigglesWarningPopup);
    end
  end

  return;
end

--
-- Function called when the Niggles EM Set popup is loaded
--
function NigglesEMSetPopupOnLoad(self)
  -- Local Variables
  local scrollFrame = self.scrollFrame

  -- Initialise some variables
  self.processEvent = true;
  self.setName      = nil;

  -- Set the frame's script handlers
  self:SetScript("OnShow", equipSetPopupOnShow);
  self:SetScript("OnEvent", equipSetPopupOnEvent);

  -- Register the frame for dragging
  self:RegisterForDrag("LeftButton");

  -- Initialise the scroll frame
  scrollFrame.update = emSetPopupListUpdate;
  scrollFrame.scrollBar.doNotHide = 1;
  HybridScrollFrame_OnLoad(scrollFrame);
  HybridScrollFrame_CreateButtons(scrollFrame, "NigglesEMSetBtnTemplate");

  -- Work around a problem with the first button shifting left-right
  scrollFrame.buttons[1]:SetWidth(scrollFrame.buttons[1]:GetWidth()-1);

  return;
end

--
-- Function called when a button in the EM Set scroll frame is
-- clicked.
--
function NigglesEMSetBtnOnClick(self, button, down)
  -- Local Variables
  local _;
  local activeSpecId = validSpecs[GetSpecialization(false, false)];
  local isActive = false;

  -- Check the button has info associated with it
  if ((self.info ~= nil) and (self.setName ~= "")) then
    -- Save the info for the selected button
    NigglesEMSetPopup.selectedInfo = self.info;

    -- Check if the item the button is for is already active
    if (NigglesEMSetPopup.mode == "set") then
      _, _, isActive = GetEquipmentSetInfoByName(self.info.name);
    elseif (NigglesEMSetPopup.mode == "spec") then
      isActive = (activeSpecId == self.info.id);
    end
    NigglesEMSetPopup.okay:SetEnabled(not isActive);

    -- Update the scroll frame
    emSetPopupListUpdate();
  end

  return;
end

--
-- Function called when a button in the EM Set scroll frame is
-- double-clicked.
--
function NigglesEMSetBtnOnDoubleClick(self, button, down)
  -- Local Variables
  local activeSpecId = validSpecs[GetSpecialization(false, false)];
  local script = NigglesEMSetPopup.okay:GetScript("OnClick");

  -- Check the button has a set name associated with it
  if ((self.info ~= nil) and
      (self.info.id ~= activeSpecId) and
      (script ~= nil)) then
    script(NigglesEMSetPopup.okay, button);
  end

  return;
end

--
-- Function called when the mouse enters a button in the EM Set
-- scroll frame.
--
function NigglesEMSetBtnOnEnter(self)
  -- Display a tooltip for the button, if possible
  if ((NigglesEMSetPopup.mode == "set") and
      (self.info ~= nil) and
      (self.info.name ~= "")) then
    GameTooltip_SetDefaultAnchor(GameTooltip, self);
    GameTooltip:SetEquipmentSet(self.info.name);
  end

  return;
end

--
-- Function called when the 'Okay' button is clicked in the EM popup
--
function NigglesEMSetPopupOkayOnClick(self, button)
  -- Local Variables
  local selectedInfo = NigglesEMSetPopup.selectedInfo;

  -- Check there is a selected set
  if (type(selectedInfo) == "table") then
    -- Make sure any event isn't processed
    NigglesEMSetPopup.processEvent = false;

    -- Equip the set or activate the spec
    if (NigglesEMSetPopup.mode == "set") then
      -- Equip the button's equipment set
      PlaySound(SOUNDKIT.IG_CHARACTER_INFO_TAB);
      EquipmentManager_EquipSet(selectedInfo.name);
    elseif (NigglesEMSetPopup.mode == "spec") then
      NigglesEMSetPopup:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED", "player");
      SetSpecialization(validSpecs[selectedInfo.id], false);
    end

    -- Hide the EM Set popup
    NigglesEMSetPopup:Hide();
  end

  return;
end

--
--
-- Function to enable/disable the EM Set spec functionality
--
function NigglesEMSetSpecEnable()
  -- Local Variables
  local _;
  local child;
  local childWidth;
  local dialog = GearManagerDialogPopup;
  local numSpecs = GetNumSpecializations(false, false);
  local heightAdj = 70;
  local scrollFrame = GearManagerDialogPopupScrollFrame;
  local specIcon;

  -- Check if the EM Set spec feature is enabled
  if ((NigglesDBC.settings.UiEMSetSpec) and
      (equipSetSpecInfo.textures[1] == nil)) then
    -- Create textures to cater for the increased size of the dialog
    child = dialog.BorderBox:CreateTexture(nil, "ARTWORK", nil);
    child:SetAtlas("macropopup-topleft", false);
    child:SetSize(18, 18);
    child:SetTexCoord(0, 1, 0.75, 1);
    child:SetPoint("BOTTOMLEFT", scrollFrame, "TOPLEFT", 3, 0);
    equipSetSpecInfo.textures[1] = child;
    child = dialog.BorderBox:CreateTexture(nil, "ARTWORK", nil);
    child:SetAtlas("macropopup-topright", false);
    child:SetSize(18, 18);
    child:SetTexCoord(0, 1, 0.75, 1);
    child:SetPoint("BOTTOMRIGHT", scrollFrame, "TOPRIGHT", 39, 0);
    equipSetSpecInfo.textures[2] = child;
    child = dialog.BorderBox:CreateTexture(nil, "ARTWORK", nil);
    child:SetAtlas("_macropopup-top", false);
    child:SetSize(0, 10);
    child:SetHorizTile(true);
    child:SetTexCoord(0, 1, 0.85294, 1);
    child:SetPoint("TOPLEFT", equipSetSpecInfo.textures[1], "TOPRIGHT", 0, -5);
    child:SetPoint("RIGHT", equipSetSpecInfo.textures[2], "LEFT");
    equipSetSpecInfo.textures[3] = child;

    -- Create a title for the new section
    child = dialog:CreateFontString();
    child:SetFontObject("GameFontHighlightSmall");
    child:SetText(L["ChooseSpecializations"]);
    child:SetPoint("TOPLEFT", dialog.BorderBox.topLeft, "BOTTOMLEFT", 23, 0);
    equipSetSpecInfo.fontStrings[1] = child;

    -- Create the buttons for selecting a EM set's spec
    for idx = 1, numSpecs do
      -- Create the button
      child = CreateFrame("CHECKBUTTON", nil, dialog,
        "NigglesEMSetSpecBtnTemplate", idx);
      childWidth = child:GetWidth()+5;
      child:SetPoint("LEFT", dialog, "CENTER",
        ((idx-1)*childWidth)-(((numSpecs*childWidth)-5)/2), 0);
      child:SetPoint("TOP", equipSetSpecInfo.fontStrings[1], "BOTTOM", 0, -3);
      child:SetID(idx);
      child.specId, _, _, specIcon = GetSpecializationInfo(idx, false, false);
      child:SetNormalTexture(specIcon);
      child:SetScript("OnEnter", emSetSpecBtnOnEnter);
      equipSetSpecInfo.buttons[idx] = child;
    end

    -- Hook the dialog's event handlers
    hooksecurefunc("RecalculateGearManagerDialogPopup",
      equipSetSpecRecalculate);
    GearManagerDialogPopupOkay:HookScript("PreClick", equipSetSpecOkayPreClick);
    GearManagerDialogPopupOkay:HookScript("OnClick",  equipSetSpecOkayOnClick);
  end

  -- Adjust the height of the dialog
  dialog:SetHeight(equipSetSpecInfo.dialogHeight+
   (NigglesDBC.settings.UiEMSetSpec and heightAdj or 0));

  -- Show/Hide the frames
  for _, current in pairs(equipSetSpecInfo.textures) do
    current:SetShown(NigglesDBC.settings.UiEMSetSpec);
  end
  for _, current in pairs(equipSetSpecInfo.fontStrings) do
    current:SetShown(NigglesDBC.settings.UiEMSetSpec);
  end
  for _, current in pairs(equipSetSpecInfo.buttons) do
    current:SetShown(NigglesDBC.settings.UiEMSetSpec);
  end

  -- (Un)register for events
  if (NigglesDBC.settings.UiEMSetSpec) then
    NigglesEMSetPopup:RegisterEvent("EQUIPMENT_SWAP_FINISHED");
    NigglesEMSetPopup:RegisterEvent("EQUIPMENT_SETS_CHANGED");
    NigglesEMSetPopup:RegisterUnitEvent("UNIT_SPELLCAST_SUCCEEDED",
      "player");
    NigglesEMSetPopup:RegisterUnitEvent("UNIT_SPELLCAST_STOP",
      "player");
  else
    NigglesEMSetPopup:UnregisterEvent("EQUIPMENT_SWAP_FINISHED");
    NigglesEMSetPopup:UnregisterEvent("EQUIPMENT_SETS_CHANGED");
    NigglesEMSetPopup:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED");
    NigglesEMSetPopup:UnregisterEvent("UNIT_SPELLCAST_STOP");
  end
  if ((NigglesDBC.settings.UiLowLevelItems) or
      (NigglesDBC.settings.WarningLowLevelItems) or
      (NigglesDBC.settings.UiEMSetSpec)) then
    NigglesEMSetPopup:RegisterEvent("PLAYER_EQUIPMENT_CHANGED");
  else
    NigglesEMSetPopup:UnregisterEvent("PLAYER_EQUIPMENT_CHANGED");
  end

  -- Scan the equipped items
  equipmentScan("Character");
  if (InspectFrame ~= nil) then
    inspectedUnitUpdateIlvl();
  end

  return;
end
